import Box2D
import random
from game.entity.entmelee import EntMelee
from game.entity.entsimpleprojectile import EntSimpleProjectile

class GameMap(object):
    """
    Representation of one map in the game world.
    A map is mainly composed of:
        - a grid and a dictionary that represent the 'tiles' that define the map's floor and walls
        - a list of GameEntity objects that are in the map
    """
    
    def __init__(self, model, grid=[], tile_data={}):
        """
        Requires GameEngine as model.
        grid should look like:
            [['#', '#', '#', '#', '#', '#', '#'],
             ['#', '.', '.', '.', '.', '.', '#'],
             ['#', '.', '.', '.', '.', '.', '#'],
             ['#', '.', '.', '.', '.', '.', '#'],
             ['#', '.', '.', '.', '.', '.', '#'],
             ['#', '.', '.', '.', '.', '.', '#'],
             ['#', '#', '#', '#', '#', '#', '#']]
            The grid is accessed row first then column, i.e. grid[y][x],
            to match the visual look of the grid (first index gives row, second gives col)
            (currently only supports one tile character per co-ordinate)
        
        tile_data should look like:
            {'.' : {'name' : "floor", 'blocks_movement' : 0},
             '#' : {'name' : "wall", 'blocks_movement' : 1}}
            (data related to images and drawing is handled elsewhere)
        """
        self.model = model
        self.grid_height = len(grid)
        self.grid_width = len(grid[0])
        self._grid = grid
        self._tile_data = tile_data
        self._entities = []
        self.spawn_time = 500
        self.spawn_amount = 50
        self.spawn_chance = 0.5
        self._curr_spawn_time = self.spawn_time
        self._dead_ents = {}

    def time_update(self, time_passed):
        self._update_entities(time_passed)
        self._do_spawning(time_passed)
        # Delay corpse removal
        for ent in self._dead_ents:
            self._dead_ents[ent] += time_passed
        
    def _do_spawning(self, time_passed):
        if self.spawn_amount <= 0: return
        self._curr_spawn_time -= time_passed
        if self._curr_spawn_time <= 0:
            self._curr_spawn_time = self.spawn_time
            if random.random() <= self.spawn_chance:
                self.spawn_amount -= 1
                sx = 2 + random.random()*(self.map_size[0] - 2)
                sy = 2 + random.random()*(self.map_size[1] - 2)
                # obviously this will explode if the map is too small....
                while Box2D.b2DistanceSquared(self.model.pchar.loc, (sx, sy)) < 6**2:
                    sx = 2 + random.random()*(self.map_size[0] - 2)
                    sy = 2 + random.random()*(self.map_size[1] - 2)
                self.model.add_entity(EntMelee(), self, (sx, sy))
        
    def _update_entities(self, time_passed):
        for ent in self.all_entities:
            ent.time_update(time_passed)
            if ent.dead and not self._dead_ents.has_key(ent):
                self._dead_ents[ent] = 0 if not isinstance(ent, EntSimpleProjectile) else 1501
                continue
            # hacky AI for EntMelee :(
            if ent.name == "Melee Mob" and not ent.dead:
                distsq = Box2D.b2DistanceSquared( self.model.pchar.loc,
                                                  ent.loc )
                if distsq < 16**2 and not self.model.pchar.dead:
                    wvec = Box2D.b2Vec2(self.model.pchar.loc[0] - ent.loc[0], self.model.pchar.loc[1] - ent.loc[1])
                    ent.dir_vec = ent.phys_body.GetLocalVector(wvec)
                else:
                    vec = ent.dir_vec
                    if random.random() < 0.01:
                        vec = random.choice([Box2D.b2Vec2(0, 0), 
                                             Box2D.b2Vec2(random.randint(-1, 1), random.randint(-1, 1))
                                             ])
                    ent.dir_vec = vec
        # Remove dead stuff
        remove = []
        for ent in self._dead_ents:
            if self._dead_ents[ent] > 1500:
                remove.append(ent)
        for ent in remove:
            self.model.remove_entity(ent, self)
            del self._dead_ents[ent]
        
    @property
    def map_size(self):
        return (self.grid_width * self.model.TILE_SPACE, 
                self.grid_height * self.model.TILE_SPACE)
        
    def tile_at(self, x, y):
        """
        Returns the tile at grid-coordinate (x, y) with form
            {'name' : "floor", 'blocks_movement' : 0}
        """
        return self._tile_data[self._grid[y][x]]
    
    @property  
    def all_entities(self):
        return self._entities[:]
    
    def add_entity(self, entity, loc):
        """
        Use GameEngine.add_entity or physics won't line up.
        Add GameEntity to map at given loc.
        """
        entity.loc = loc
        entity.curr_map = self
        self._entities.append(entity)
    
    def remove_entity(self, entity):
        """
        Use GameEngine.remove_entity or physics won't line up.
        Remove given GameEntity from this map.
        """
        entity.curr_map = None
        try:
            self._entities.remove(entity)
        except ValueError:
            raise ValueError("remove_entity: Tried to remove entity when entity was not in the map.")
